// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include the GUI:
#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

// Include functions to display Mat4f and Vec4f in the GUI:
# include <imgui_cg1_helpers.h>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include GLFW/OpenGL:
#include <GLFW/glfw3.h>

// Include IO utilities:
#include <iostream>

// Include Vec4f class:
#include "src/math/Vec4f.h"

// Include Mat4f class:
#include "src/math/Mat4f.h"

// Include our shader wrapper class:
#include "src/Shader.h"

// Include our mesh class:
#include "src/Mesh.h"

// Include our new SpinningTop class:
#include "src/game/SpinningTop.h"

/**
 * Initializes the application including the GUI
 */

int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(860, 560, "Computergrafik 1", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls

    // Setup Dear ImGui style:
    ImGui::StyleColorsDark();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Defines the with of the left menu:
    int menuWidth = 260;

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 16.0f);

    // Enable depth test:
    glEnable(GL_DEPTH_TEST);

    // Should use perspective projection?:
    bool usePerspectiveTransformation = true;

    // Load meshes:
    Mesh groundMesh(CMAKE_SOURCE_DIR "/models/World.obj");
    Mesh spinningTopMesh(CMAKE_SOURCE_DIR "/models/SpinningTop.obj");

    // Load shader:
    Shader groundShader(CMAKE_SOURCE_DIR "/shader/world.vert", CMAKE_SOURCE_DIR "/shader/world.frag");
    Shader spinningTopShader(CMAKE_SOURCE_DIR "/shader/spinningTop.vert", CMAKE_SOURCE_DIR "/shader/spinningTop.frag");

    // We define a world which has a groundMesh and multiple spinning tops in it:
    World world(groundMesh, groundShader);

    // Add some spinning tops to the world:
    world.spinningTops.emplace_back(SpinningTop(spinningTopMesh, spinningTopShader, Vec4f(0.7f,0.0f,-1.7f)));
    world.spinningTops.emplace_back(SpinningTop(spinningTopMesh, spinningTopShader, Vec4f(1.3f,0.0f,-1.7f)));
    world.spinningTops.emplace_back(SpinningTop(spinningTopMesh, spinningTopShader, Vec4f(0.0f,0.0f,-4.5f)));

    // View
    int currentViewSelection = 0;
    const char* viewSelectionNames[] = {"Übersicht", "Kreisel 1", "Kreisel 2", "Kreisel 3"};

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Subtract right menu:
        display_w -= menuWidth;

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();

            // Create Window with settings:
            ImGui::SetNextWindowPos(ImVec2(0,0));
            ImGui::SetNextWindowSize(ImVec2(float(menuWidth),float(display_h)));
            ImGui::Begin("Einstellungen", nullptr, ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoCollapse | ImGuiWindowFlags_NoResize);
            ImGui::Combo("Ansicht", &currentViewSelection, &viewSelectionNames[0], IM_ARRAYSIZE(viewSelectionNames));
            ImGui::Separator();

            ImGui::Checkbox("Perspektivische Transformation", &usePerspectiveTransformation);
            ImGui::End();
        }

        // Setup the GL viewport
        glViewport(menuWidth, 0, display_w, display_h);
        glClearColor(0.6f, 0.725f, 0.8f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Calculate aspect ratio (might be used for perspective transformation):
        float aspectRatio = float(display_w) / display_h;

        // Ensure that both sides of the triangle are rendered:
        glEnable(GL_CULL_FACE);
        glCullFace(GL_FRONT);

        // Initially, we simply have an identity matrix for the projection, so that it
        // does not affect the position when it is multiplied with other matrices:
        Mat4f projection;

        // Only when perspective transformation is wanted, it's written
        // into the projection matrix:
        if(usePerspectiveTransformation){
            projection = Mat4f::perspectiveTransformation(aspectRatio);
        }

        Mat4f view;

        // If selection is "Uebersicht", then define a fix view matrix:
        if(currentViewSelection == 0){
            view = Mat4f::rotationX(-0.8f) * Mat4f::translation(0,-7.5, 14);
        } else {
            // If we want to render out of the perspective of one of the spinning tops,
            // we can use the inverse of their model matrix (and translate it a little bit
            // so the view is not in the middle of the spinning top, but behind it:
            view = Mat4f::translation(0.f,-0.5f,2.f) * world.spinningTops[currentViewSelection-1].getModelMatrix().inverse();
        }

        // Render the world and simulate and render the spinning tops:
        world.render(view, projection);

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}

